# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 25.10.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Functions for Principal Component Analysis, and Fisher's Linear Discriminan Analysis.
"""

# pylint: disable=bad-indentation

import numpy as np
from sklearn.decomposition import PCA # pylint: disable=import-error

def pca(feature_images):
  """Computes the PC of every pixel, and returns them.

  Args:
    feature_images: Numpy array of shape [num_features, height, width].
  Returns:
    principal_images: Numpy array of shape [num_features, height, width], where
                      each channel correspond to a principal component. It is
                      sorted in descending order of principality (first
                      principal component in channel 0 etc).
    eigvals: Sorted eigenvalues of the feature covariance matrix computed from
             the data in the feature images.
  """

  num_features, height, width = feature_images.shape
  data = np.zeros((num_features, height*width))
  for channel in range(num_features):
    data[channel, :] = feature_images[channel, :, :].flatten()

  covmat = np.cov(data)
  eigvals, eigvecs = np.linalg.eig(covmat)

  # Sort eigenvalues and eigenvectors by eigenvalues in descending order
  idx = eigvals.argsort()[::-1]
  eigvals = eigvals[idx]
  eigvecs = eigvecs[:, idx]

  pc_data = np.dot(eigvecs.T, data)

  principal_images = np.zeros((num_features, height, width))
  for channel in range(num_features):
    principal_images[channel, :, :] = pc_data[channel, :].reshape((height, width))

  # Test implementation against sklearn
  skl_pca = PCA()
  skl_pc_data = skl_pca.fit_transform(data.T).T

  skl_principal_images = np.zeros((num_features, height, width))
  for channel in range(num_features):
    skl_principal_images[channel, :, :] = skl_pc_data[channel, :].reshape((height, width))

  #print(np.linalg.norm(principal_images - skl_principal_images))
  # NOTE: There seem to be a constant difference between the two, but I cannot
  # figure out why this is so at the moment

  return principal_images, eigvals


def lda(feature_images, class_vectors):
  """Computes the PC of every pixel, and returns them.

  Args:
    feature_images: Numpy array of shape [num_features, height, width].
    class_vectors: List of numpy arrays, one for each class, where each array
                   is of shape [num_features, num_samples_in_class], and
                   num_samples_in_class is specific to each class.
  Returns:
    ld_class_vectors: List like class_vectors of transformed feature vectors.
    eigvals: Sorted eigenvalues of the inner product of the inverse of the
             within-class scatter matrix and the between-class scatter matrix.
  """

  num_features, height, width = feature_images.shape
  num_classes = len(class_vectors)

  class_data = np.concatenate(class_vectors, axis=1)

  data = np.zeros((num_features, height*width))
  for channel in range(num_features):
    data[channel, :] = feature_images[channel, :, :].flatten()

  overall_mean = np.mean(class_data, axis=1)
  class_means = np.zeros((num_features, num_classes))
  for c in range(num_classes):
    class_means[:, c] = np.mean(class_vectors[c], axis=1)

  within_class_scatter_matrix = np.zeros((num_features, num_features))
  for c in range(num_classes):
    num_dp_in_class = class_vectors[c].shape[1]
    within_class_scatter_matrix += np.cov(class_vectors[c])*(num_dp_in_class - 1)

  # This is equivalent, but orders of magnitude slower
  #within_class_scatter_matrix_2 = np.zeros((num_features, num_features))
  #for c in range(num_classes):
  #  temp_class_scatter_matrix = np.zeros((num_features, num_features))
  #  for feature_vector in class_vectors[c].T:
  #    mean_diff = (feature_vector - class_means[:, c]).reshape((num_features, 1))
  #    temp_class_scatter_matrix += np.dot(mean_diff, mean_diff.T)
  #  within_class_scatter_matrix_2 += temp_class_scatter_matrix
  #
  #np.testing.assert_array_almost_equal(within_class_scatter_matrix,
  #                                     within_class_scatter_matrix_2)

  between_class_scatter_matrix = np.zeros((num_features, num_features))
  for c in range(num_classes):
    num_dp_in_class = class_vectors[c].shape[1]
    mean_diff = (class_means[:, c] - overall_mean).reshape((num_features, 1))
    between_class_scatter_matrix += num_dp_in_class * np.dot(mean_diff, mean_diff.T)

  prodmat = np.dot(np.linalg.inv(within_class_scatter_matrix), between_class_scatter_matrix)

  eigvals, eigvecs = np.linalg.eig(prodmat)
  idx = eigvals.argsort()[::-1]
  eigvals = eigvals[idx]
  eigvecs = eigvecs[:, idx]

  ld_data = np.dot(eigvecs.T, data)
  ld_images = np.zeros((num_features, height, width))
  for channel in range(num_features):
    ld_images[channel, :, :] = ld_data[channel, :].reshape((height, width))

  ld_class_data = np.dot(eigvecs.T, class_data)
  ld_class_vectors = []
  prev_ind = 0
  for c in range(num_classes):
    next_ind = prev_ind + class_vectors[c].shape[1]
    ld_class_vectors.append(ld_class_data[:, prev_ind:next_ind])
    prev_ind = next_ind

  return ld_images, ld_class_vectors, eigvals
